/** File:      mucalculus.P --- Encoding of SOS semantics of modal mu-calculus
** Author(s): C.R. Ramakrishnan
** Contact:   lmc@cs.sunysb.edu
** 
** Copyright (C) SUNY at Stony Brook, 1998-2000
** 
** XMC is free software; you can redistribute it and/or modify it under the
** terms of the GNU Library General Public License as published by the Free
** Software Foundation; either version 2 of the License, or (at your option)
** any later version.
** 
** XMC is distributed in the hope that it will be useful, but WITHOUT ANY
** WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
** FOR A PARTICULAR PURPOSE.  See the GNU Library General Public License for
** more details.
** 
** You should have received a copy of the GNU Library General Public License
** along with XMC; if not, write to the Free Software Foundation,
** Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
**
** $Id: mucalculus.P,v 1.1.1.1 2002/09/04 19:54:55 mmc Exp $
** 
*/

:- import sk_not/1 from tables.

:- import format/2 from format.
:- import length/2 from basics.
:- import ith/3 from basics.

%% rec_models(S, F):  Does state S model formula with name F?
:- table rec_models/2.

rec_models(State_s, X) :-
	fDef(X, Y), models(State_s, Y).

%% models(S, X): Does state S model formula expression X?
:- index models/2-2.

%% Case split on formula expression:

% True
models(_State_s, tt).

% Recursion
models(State_s, form(X)) :-
	rec_models(State_s, X).

% Conjunction
models(State_s, fAnd(X_1, X_2))	:-
	models(State_s, X_1), models(State_s, X_2).

% Disjunction
models(State_s, fOr(X_1, X_2))	:-
	models(State_s, X_1); models(State_s, X_2).

% Diamond: <A>_p
sum([], P) :-
	P is 0.
sum([H|L], P) :-
	sum(L, P1),
	P is H + P1.

check_bound(Prob, ge(Bound)) :-
	Prob >= Bound.
check_bound(Prob, gt(Bound)) :-
	Prob > Bound.
check_bound(Prob, le(Bound)) :-
	Prob =< Bound.
check_bound(Prob, lt(Bound)) :-
	Prob < Bound.

models(State_s, fProbDiam(Act_a, Bound, X)) :-
	transition(State_s, PSteps),
	findall(Prob, check_prob_diam(PSteps, Act_a, Prob, X), ProbList),
	sum(ProbList, TotalProb),
	format("<~w>~w: ~w~n", [Act_a, X, TotalProb]),
	(nonvar(Bound) -> check_bound(TotalProb, Bound)
		       ;  Bound = TotalProb).

check_prob_diam(PSteps, Act_a, Prob, X) :-
	length(PSteps, L),
	upto(L, I),
	ith(I, PSteps, pstep(BProb, Act_b, State_t)),
	(comp(Act_a, Act_b) -> models(State_t, X),
			      Prob = BProb
			   ;  (Act_b == tau -> models(State_t, fProbDiam(Act_a, TProb, X)),
					       Prob is BProb * TProb
					    ;  Prob = 0)).

% Reachability: (True Until <A>X)_p
models(State_s, fProbReach(Act_a, Bound, X)) :-
	transition(State_s, PSteps),
	findall(Prob, check_prob_reach(PSteps, Act_a, Prob, X), ProbList),
	sum(ProbList, TotalProb),
	format("EF(<~w>~w): ~w~n", [Act_a, X, TotalProb]),
	(nonvar(Bound) -> check_bound(TotalProb, Bound)
		       ;  Bound = TotalProb).

check_prob_reach(PSteps, Act_a, Prob, X) :-
	length(PSteps, L),
	upto(L, I),
	ith(I, PSteps, pstep(BProb, Act_b, State_t)),
	(comp(Act_a, Act_b) -> models(State_t, X),
			       Prob = BProb
			    ;  models(State_t, fProbReach(Act_a, TProb, X)),
			       Prob is BProb * TProb).

comp(in(C, V), in(D, V)) :- C==D.
comp(out(C, V), out(D, V)) :- C==D.
comp(outbound(C, V), outbound(D, V)) :- C==D.
comp(outbound1(C, V), outbound1(D, V)) :- C==D.
comp(tau, tau).

/*
% Diamond Minus: <-A>
models(State_s, fDiamMinus(Act_a, X)) :-
%	trans_not_a(State_s, Act_a, State_t),
	transition(State_s, Act_b, State_t),
	\+ comp(Act_a, Act_b),
	models(State_t, X).

% Diamond Set: <{S}>
models(State_s, fDiamSet(Set_a, X)) :-
	trans_member(State_s, Set_a, State_t),
        models(State_t, X).

% Diamond Set Minus: <-{S}>
models(State_s, fDiamSetMinus(Set_a, X)) :-
	trans_not_member(State_s, Set_a, State_t),
	models(State_t, X).

% Box: [A]
models(State_s, fBox(Act_a, X))	:-
	forall(State_t,
	       trans_a(State_s, Act_a, State_t),
	       models(State_t, X)).

% Box Minus: [-A]
models(State_s, fBoxMinus(Act_a, X)) :-
	forall(State_t, 
	       trans_not_a(State_s, Act_a, State_t),
	       models(State_t, X)).

% Box Set: [{S}]
models(State_s, fBoxSet(Set_a, X)) :-
        forall(State_t, 
	       trans_member(State_s, Set_a, State_t),
	       models(State_t, X)).

% Box Set Minus: [-{S}]
models(State_s, fBoxSetMinus(Set_a, X)) :-
	forall(State_t,
	       trans_not_member(State_s, Set_a, State_t),
	       models(State_t, X)).

% Negation
models(State_s, neg_form(X)) :-
	sk_not(rec_models(State_s, X)).
*/
/*
    Auxilliary trans relations:
        transition(S, A, T): there is a transition A from S to T this relation is defined in calculus.P
   
        trans_not_a(S, A, T): there is a non A transition from S to T.
   
        trans_member(S, L, T): there is a transition A from S to T such that A is in the set L.
   
        trans_not_member(S, L, T): there is a transition A from S to T such that A is not in the set L.
*/
trans_a(State_s, Act_a, State_t) :-
        transition(State_s, Act_b, State_t),
        comp(Act_b, Act_a).  

trans_not_a(State_s, Act_a, State_t) :-
	transition(State_s, Act_b, State_t),
%	\+ (Act_b == Act_a).
	\+ comp(Act_b, Act_a).

trans_member(State_s, Set_a, State_t) :-
        transition(State_s, Act_b, State_t),
        member(Act_b, Set_a).

trans_not_member(State_s, Set_a, State_t) :-
	transition(State_s, Act_b, State_t),
	\+ member(Act_b, Set_a).

%% ----------------------------------------------------------------------

	% forall substitutions of Variables,
	% whenever Antecedent is true, so is Consequent

forall(_Variables, Antecedent, Consequent) :-
	findall(Consequent, Antecedent, AllConsequents),
	all_true(AllConsequents).

all_true([]).
all_true([C|Cs]) :- call(C), all_true(Cs).

% member(X, [H|T]) :- X = H ; member(X, T).
member(X, [H|T]) :- comp(X, H); member(X, T).

:- import trie_intern/3, trie_interned/3 from intern.

/*intern(Term, ID) :-
        (var(Term)
                ->  trie_interned(Term, ID, _)
                ;   (var(ID)
                        -> trie_intern(Term, ID, _)
                        ; trie_interned(Term, NewId, _), NewId = ID
                )
        ).*/

intern(Term, ID, Var) :-
        (var(Term)
                ->  trie_interned(Term, ID, Var)
                ;   (var(ID)
                        -> trie_intern(Term, ID, Var)
                        ; trie_interned(Term, NewId, Var), NewId = ID
                )
        ).


start(P, L) :-
	L = proc(P).

transition(S1, PSteps) :-
        trans(S1, PSteps, _, _).

/*start(P, L) :-
 	P1 = proc(P),	
	intern(P1,L).

transition(L1,A,L2) :-
	intern(S1, L1),
	trans(S1,A,S2),
	intern(S2, L2).*/

/*start(P, f(L, Var)) :-
        P1 = proc(P),
        intern(P1,L,Var).
 
transition(f(L1, Var1), A, f(L2, Var2)) :-
        intern(S1, L1, Var1),
        trans(S1,A,_,S2,0,_),
        intern(S2, L2, Var2).*/

