% To derive a symbolic transition graph for a specific process.
% Commands: 
%   stg(Proc). - print out the graph of the process Proc.
%   stg_log(Proc). - output the graph of the process Proc into a file named after the name of Proc.
%   stg_all(ProcList). - print out the graphs of processes in the list ProcList.
%   stg_all_log(ProcList). - output the graphs of processes in the list ProcList into a file named "a.stg".
%   stg_all_log(ProcList, FName). - output the graphs of processes in the list ProcList into a file named after FName.
%
% Note that the parameters of the process should be given as atoms, namely written with a lower-case first letter.

:- compiler_options([xpp_on]).
#include "standard.h"

:- import format/2 from format.

:- table ref/2.

:- dynamic file_no/1.
:- dynamic save/2.
:- dynamic pedge_num/1.
:- dynamic node_num/1.
:- dynamic last_node_num/1.
:- dynamic edge_num/1.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% print predicates
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
print_listno([], _).
print_listno([H|L], N) :-
	format("[~w: ~w]~n", [N, H]),
	N1 is N + 1,
	print_listno(L, N1).

print_node(N, S) :-
	format("#~w: ~w~n", [N, S]).

print_matches([]):-
	format("~n", []).
print_matches([_, matchf(K)|L]) :-
	format("~nReading the file ~w~n", [K]),
	open(file(K), read, F),
	file_read_line_list(F, S),
	close(F),
	format("~s.", [S]),
	print_matches(L).
print_matches([X, mismatch(Y)|L]) :-
	format("~w != ~w. ", [X, Y]),
	print_matches(L).

print_edge(N, A, V, M) :-
	set_edge_num(K),
	format("*~w: ~w -- [", [K, N]),
	print_matches(V),
	format("], ~w --> ~w~n", [A, M]).

print_nodes([]).
print_nodes([(N, S)|T]) :-
	print_node(N, S), print_nodes(T).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% index predicates
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
exit(S) :-
	node_num(N),
	edge_num(M),
	format("~nError at ~w after ~w nodes and ~w edges!~n", [S, N, M]),
	halt.

set_node_num(N) :-
	node_num(K),
	N is K + 1,
	retract(node_num(K)),
	assert(node_num(N)).

set_edge_num(N) :-
	edge_num(K),
	N is K + 1,
	retract(edge_num(K)),
	assert(edge_num(N)).

set_pedge_num(N) :-
	pedge_num(K),
	N is K + 1,
	retract(pedge_num(K)),
	assert(pedge_num(N)).
	
get(S, N) :-
	save(S, I) -> N = I; N = -1.
	
set(S, N) :-
	set_node_num(N),
	print_node(N, S),
	assert(save(S, N)).

ref(S, N) :-
        nonvar(S) -> (
	get(S, M),
	integer(M) -> ((M == -1) -> set(S, N); N = M)
		;
		exit(S)).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% name predicates
% atom: a
% var: V, _a
% compound: name(V), name(_a)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

:- table free_name/1.

:- dynamic fname/1.
:- dynamic bname_num/1.

free_name(X) :-
	atom(X) -> assert(fname(X))
		;
		   (compound(X) -> (functor(X, F, N),
		   		   (F == name; F == matchf; N == 0;
				   (upto(N,I), 
		   		    arg(I,X,X1), 
		    		    free_name(X1))))
				;
				   var(X)).

set_bname_num(N) :-
	bname_num(K),
	N is K + 1,
	retract(bname_num(K)),
	assert(bname_num(N)).

bound_name(X) :-
	var(X) -> (set_bname_num(K), format(">~w: ~w~n", [K, X]))
		;
		  (compound(X) -> (functor(X, _, N),
				   upto(N,I), 
		   		   arg(I,X,X1), 
		    		   bound_name(X1))
			       ;
			          atom(X)).

get_name(tau).

get_name(out(X, Y)) :-
	free_name(X), findall(_, free_name(Y), _).

get_name(in(X, Y)) :-
	free_name(X), findall(_, bound_name(Y), _).

%get_name(outbound(X, name(Y))) :-
%	free_name(X), bound_name(Y).

%get_name(outbound1(X, name(Y))) :-
%	free_name(X), bound_name(Y).

get_name(outbound(X, Y)) :-
	free_name(X), findall(_, bound_name(Y), _).

get_name(outbound1(X, Y)) :-
	free_name(X), findall(_, bound_name(Y), _).

get_name([X|L]) :-
	free_name(X), get_name(L).

get_name([]).

print_name(F) :-
	findall(FN, fname(FN), FNames), length(FNames, F),
	print_listno(FNames, 1).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% main
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

:- import length/2 from basics.
:- import append/3 from basics.
:- table go2/1.

rec_go2([]).
rec_go2([H|L]) :-
	go2(H);
	rec_go2(L).

go2(Curr) :-
	ref(Curr, N),
	trans(Curr, PSteps, _, V),
	get_name(V),
	set_pedge_num(K),
%	format("*~w: ~w == ~w~n", [K, N, V]),
	format("*~w: ~w == ", [K, N]),
	print_matches(V),
	next(PSteps, []),
	fail.

next([], Nexts) :-
	rec_go2(Nexts).

next([pstep(Prob, A, Next)|Others], Nexts) :-
	ref(Next, M),
	get_name(A),
	set_edge_num(K),
	format("'~w: -- '~w':~w --> ~w~n", [K, Prob, A, M]),
	append(Nexts, [Next], NewNexts),
	next(Others, NewNexts).

reset :-
	retractall(edge_num(_)), retractall(pedge_num(_)),
	retractall(fname(_)), retractall(bname_num(_)),
	retractall(file_no(_)),
	assert(edge_num(0)), assert(pedge_num(0)),
	assert(bname_num(0)),
	assert(file_no(0)),
	abolish_table_pred(free_name/1),
	abolish_table_pred(ref/2),
	abolish_table_pred(go2/1).

clear :-
	retractall(save(_,_)),
	retractall(node_num(_)), assert(node_num(0)), 
	retractall(last_node_num(_)), assert(last_node_num(0)),
	reset.

clearmfile :-
	writeln('Deleting the temporary files...'),
	file_no(K),
	(K == 0; 
	findall(_, (upto(K, I),
		    fmt_write_string(S, ".m%d", args(I)),
		    path_sysop(rm, S)), _)),
	writeln('Deleting the temporary files...OK').

info(Proc, F) :-
	last_node_num(N0), node_num(N), edge_num(M), pedge_num(K), bname_num(B),
	N1 is N - N0,
	format("~n+++ Statistics of ~w +++~n(New) Nodes: ~w~nEdges: ~w~nP-Steps: ~w~nFree Names: ~w~nBound Names: ~w~n~n", [Proc, N1, M, K, F, B]).

explore(Proc, F) :-
	clear,
	go2(proc(Proc)); % findall((Y, X), save(X, Y), Nodes), print_nodes(Nodes),
	print_name(F).
	
stg(Proc) :-
	explore(Proc, F),
	info(Proc,F),
	clearmfile.

stg_log(Proc) :-
	functor(Proc, PName, _),
	open(file(PName), write, FP),
	set_output(FP),
	stg(Proc),
	set_output(STDOUT),
	close(FP).

explore_all([]).

explore_all([H|L]) :-
	retractall(last_node_num(_)), node_num(N), assert(last_node_num(N)),
	reset,
	go2(proc(H));
	print_name(F),
	info(H, F),
	clearmfile,
	explore_all(L).

stg_all(ProcList) :-
	retractall(save(_,_)),
	retractall(node_num(_)), assert(node_num(0)), 
	explore_all(ProcList).

stg_all_log(ProcList, FName) :-
	open(file(FName), write, FP),
	set_output(FP),
	stg_all(ProcList),
	set_output(STDOUT),
	close(FP).

stg_all_log(ProcList) :-
	stg_all_log(ProcList, 'a.stg').
